/*
 * tmos - aduc841.c
 * 	version 0.0.0
 * 
 * Copyright (C) 2009  Marek Piwonski <marek@microster.pl>
 * Copyright (C) 2009, 2010  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <tmos/kernel.h>
#include <tmos/adc.h>
#include <asm/aduc841.h>
#include <asm/irqs.h>

// #define ADC_DEBUG

#define ADC_TEMP		0x8
#define ADC_DAC0		0x9
#define ADC_DAC1		0xa
#define ADC_AGND		0xb
#define ADC_VREF		0xc
#define ADC_STOP		0xf

#define ADCCON1_MD1		0x80
#define ADCCON1_EXT_REF		0x40
#define ADCCON1_CK_32		0x00
#define ADCCON1_CK_4		0x10
#define ADCCON1_CK_8		0x20
#define ADCCON1_CK_2		0x30
#define ADCCON1_AQ_1		0x00
#define ADCCON1_AQ_2		0x04
#define ADCCON1_AQ_3		0x08
#define ADCCON1_AQ_4		0x0c
#define ADCCON1_T2C		0x02
#define ADCCON1_EXC		0x01

#define ADCCON2_ADCI	0x80
#define ADCCON2_DMA	0x40
#define ADCCON2_CCONV	0x20
#define ADCCON2_SCONV	0x10
#define ADCCON2_CH(x)	((x) & 0xf)

#define ADCCON3_BUSY	0x80
#define ADCCON3_AVG1	0x10
#define ADCCON3_AVG15	0x00
#define ADCCON3_AVG31	0x20
#define ADCCON3_AVG63	0x30
#define ADCCON3_CALIB	0x04
#define ADCCON3_TYPICAL	0x02
#define ADCCON3_SCAL	0x01

#define ADCCON3_TYPICAL_GAIN	ADCCON3_TYPICAL
#define ADCCON3_TYPICAL_OFFSET	0

#define RELOAD_VALUE \
	(uint16_t)(65536 - (CONFIG_OSCILLATOR / CONFIG_ADC_ADUC841_FREQ))

#ifdef CONFIG_ADC_ADUC841_USE_XDATA
#define bdata __xdata
#else
#define bdata __data
#endif

#define F_N		CONFIG_ADC_ADUC841_CHANNEL_COUNT

static uint8_t nr_ch;

#ifdef CONFIG_ADC_ADUC841_PROFILE
static tsc_t tt;
#endif

#ifdef CONFIG_ADC_SYNC
static uint8_t nr_seq;

#define adc_seq_inc()	do { nr_seq++; } while (0)

/**
 * adc_sync - wait until channel ch is ready
 * @ch: channel
 */
unsigned char adc_sync(unsigned char seq) __reentrant
{
	while (nr_seq == seq)
		cpu_relax();
	smp_rmb();
	return nr_seq;
}

/**
 * adc_wait_ch - wait until channel ch is ready
 * @ch: channel
 */
void adc_wait_ch(unsigned char ch) __reentrant
{
	while (nr_ch == ch)
		cpu_relax();
	smp_rmb();
}
#else
#define adc_seq_inc()	do { } while (0)
#endif

#define DEFINE_ADC(x) \
	void aduc_ ## x ## _process(int16_t) __using(ADUC_TIMER_BANK); \
	int16_t aduc_ ## x ## _get(void);

DEFINE_ADC(0)
DEFINE_ADC(1)
DEFINE_ADC(2)
DEFINE_ADC(3)
DEFINE_ADC(4)
DEFINE_ADC(5)
DEFINE_ADC(6)
DEFINE_ADC(7)

int16_t adc_get(unsigned int ch)
{
	int16_t ret = 0;

	switch (ch) {
	case 0:
		ret = aduc_0_get();
		break;
	case 1:
		ret = aduc_1_get();
		break;
	case 2:
		ret = aduc_2_get();
		break;
	case 3:
		ret = aduc_3_get();
		break;
	case 4:
		ret = aduc_4_get();
		break;
	case 5:
		ret = aduc_5_get();
		break;
	case 6:
		ret = aduc_6_get();
		break;
	case 7:
#ifdef CONFIG_ADC_ADUC841_PROFILE
		ret = tt;
#else
		ret = aduc_7_get();
#endif
		break;
	default:
		ret = 0;
	}
	return ret;
}

#define update_value(ch, v) \
	do { \
		switch (ch) { \
		case 0: \
			aduc_0_process(v); \
			break; \
		case 1: \
			aduc_1_process(v); \
			break; \
		case 2: \
			aduc_2_process(v); \
			break; \
		case 3: \
			aduc_3_process(v); \
			break; \
		case 4: \
			aduc_4_process(v); \
			break; \
		case 5: \
			aduc_5_process(v); \
			break; \
		case 6: \
			aduc_6_process(v); \
			break; \
		case 7: \
			aduc_7_process(v); \
			break; \
		} \
	} while (0);


#define start_conv_ch() \
	do { \
		ADCCON2 = nr_ch | ADCCON2_SCONV; \
	} while (0)

#if 0
static uint16_t adc_read(void) __using(ADUC841_ADC_BANK) __naked
{
	__asm
; high byte
		mov	a, #0x0F
		anl	a, _ADCDATAH
		add	a, #0xf8
		mov	dph, a
; low byte
		mov	a, _ADCDATAL
		mov	dpl,a
		ret
	__endasm;	
}
#else
static uint16_t adc_read(void) __using(ADUC841_ADC_BANK)
{
	uint8_t high = ADCDATAH & 0x0f;
	uint8_t low = ADCDATAL;
	uint16_t x = (high << 8) | low;

	return (int16_t) x - 2048;
}
#endif

void aduc_timer_irq(void) __interrupt(ADUC_TIMER_INTERRUPT)
		__using(ADUC_TIMER_BANK)
{
	uint16_t x;

#ifdef CONFIG_ADC_ADUC841_PROFILE
	tt = tsc_read_low();
#endif
	if (!ADCI)
		goto out;
	x = adc_read();
	update_value(nr_ch, x);

/*	smp_wmb(); */
	nr_ch++;
#if CONFIG_ADC_ADUC841_CHANNEL_COUNT == 8
	nr_ch &= 0x07;
#elif CONFIG_ADC_ADUC841_CHANNEL_COUNT == 4
	nr_ch &= 0x03;
#elif CONFIG_ADC_ADUC841_CHANNEL_COUNT == 2
	nr_ch &= 0x01;
#else
	if (nr_ch >= CONFIG_ADC_ADUC841_CHANNEL_COUNT)
		nr_ch = 0;
#endif
/*	smp_wmb(); */
	if (!nr_ch)
		adc_seq_inc();

	start_conv_ch();

#ifdef CONFIG_ADC_ADUC841_PROFILE
	tt = tsc_read_low() - tt;
#endif

out:
	TF2 = 0;	/* clear interrupt bit */
}

void adc_set_freq(unsigned long freq) __reentrant
{
	uint16_t x;

	/*
	 * 12.8 kHz is a good value for 11MHz clock. This should
	 * be configureable.
	 */
	if (freq > 12800) {
		freq = 12800;
	}

	freq = (CONFIG_OSCILLATOR + freq / 2) / freq;
	if (!freq) {
		freq = 1;
	} else if (freq >= 65536) {
		freq = 65536;
	}

	x = 65536 - freq;

	TR2 = 0;			/* stop the timer2 */
	TF2 = 0;			/* clear pending interrupt */
	RCLK = 0;			/* serial don't use this timer */
	TCLK = 0;			/* serial don't use this timer */
	CAP2 = 0;			/* autoreload mode */
	EXEN2 = 0;			/* normal reload */
	CNT2 = 0;			/* use on-chip clock source */
	TL2 = x & 0xff;			/* initial L value */
	RCAP2L = x & 0xff;		/* reload L value */
	TH2 = x >> 8;			/* initial H value */
	RCAP2H = x >> 8;		/* reload H value */
#if CONFIG_ADC_ADUC841_TIMER_HIGHPRIO
	PT2 = 1;
#endif
	ET2 = 1;			/* enable timer2 interrupt */
	TR2 = 1;			/* start the timer */
}

static inline void adc_wait(void)
{
	while (ADCCON3 & ADCCON3_BUSY);
}

static void ad_ofs_cal(void)
{
	ADCCON2 = ADCCON2_SCONV	| ADCCON2_CH(ADC_AGND);
	ADCCON3 = ADCCON3_AVG63 | ADCCON3_CALIB | ADCCON3_TYPICAL_OFFSET
		| ADCCON3_SCAL;
	adc_wait();
}

static void ad_gain_cal(void)
{
	ADCCON2 = ADCCON2_SCONV	| ADCCON2_CH(ADC_VREF);
	ADCCON3 = ADCCON3_AVG63 | ADCCON3_CALIB | ADCCON3_TYPICAL_GAIN
		| ADCCON3_SCAL;
	adc_wait();
}

static void adc_init(void)
{
/*	ADCLK = OSC/2, Aquisition time = 4 ADCLK */
	ADCCON1 = ADCCON1_MD1 | ADCCON1_CK_2 | ADCCON1_AQ_4;
	ADCCON2 = 0;
	ADCCON3 = ADCCON3_AVG1;
	EADC = 0;
	ad_ofs_cal();
	ad_gain_cal();
	adc_set_freq(CONFIG_ADC_ADUC841_FREQ);
}

device_initcall(adc_init)
