/*
 * tmos - beerbus.c
 *
 * beerbus bus controller
 *
 * Copyright (C) 2010  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * beer bus controller
 *
 * P2_7 - SYSCLK		P0_7 - AD7
 * P2_6	- CLK			P0_6 - AD6
 * P2_5	- DEVSEL#		P0_5 - AD5
 * P2_4	- CMD2			P0_4 - AD4
 * P2_3	- CMD1			P0_3 - AD3
 * P2_2	- CMD0			P0_2 - AD2
 * P2_1	- X			P0_1 - AD1
 * P2_0	- X			P0_0 - AD0
 */

/*
 * This is beerbus controller implementation. This version supports
 * only 8-bit addressing.
 */

#include <tmos/kernel.h>
#include <tmos/string.h>
#include <asm/aduc841.h>
#include <asm/system.h>
#include <asm/io.h>
#include <tmos/errno.h>

#define clk_hi()	do { P2_6 = 1; } while (0)
#define clk_lo()	do { P2_6 = 0; } while (0)
#define clk()		do { clk_lo(); clk_hi(); } while (0)

#define devsel_hi()	do { P2_5 = 1; } while (0)
#define devsel_lo()	do { P2_5 = 0; } while (0)
#define devsel_get()	(P2_5)

#define CMD_READ	0
#define CMD_WRITE	1
#define CMD_ALE		2
#define CMD_RESET	6
#define CMD_NOP		7
#define CMD_MASK	7

#define P2_MASK		0x5c

#define cmd_set(x) \
	do { \
		P2 &= ~P2_MASK; \
		P2 |= ((x & CMD_MASK) << 2) | 0xa0; \
	} while (0)

#define ad_set(x)	do { P0 = (x); } while (0)
#define ad_get()	(P0)
#define ad_free()	do { P0 = 0xff; } while (0)

/**
 * iowrite8 - write byte @val to I/O memory address @addr
 * @val: value to write
 * @addr: I/O memory address
 */
void iowrite8(u8 val, unsigned int addr) __reentrant
{
	cpu_flags_t flags;

	/*
	 * FIXME: spinlock should be used instead of disabling IRQ
	 * The code on UP will be the same.
	 */
	local_irq_save(flags);

	cmd_set(CMD_ALE);
	ad_set(addr & 0xff);
	clk_hi();

	cmd_set(CMD_WRITE);
	ad_set(val);
	clk_hi();

	cmd_set(CMD_NOP);
	ad_free();
	clk_hi();

	local_irq_restore(flags);
}

/**
 * ioread8 - read byte from I/O memory address @addr
 * @addr: I/O memory address
 */
u8 ioread8(unsigned int addr) __reentrant
{
	cpu_flags_t flags;
	u8 ret = 0xff;

	local_irq_save(flags);

	cmd_set(CMD_ALE);
	ad_set(addr & 0xff);
	clk_hi();

	cmd_set(CMD_READ);
	ad_free();
	clk_hi();


	cmd_set(CMD_NOP);
	if (!devsel_get())
		ret = ad_get();
	clk_hi();

	cmd_set(CMD_NOP);
	clk_hi();

	local_irq_restore(flags);
	return ret;
}

/**
 * ioread16 - read 16-bit word from I/O memory address @addr
 * @addr: I/O memory address
 */
u16 ioread16(unsigned int addr) __reentrant
{
	cpu_flags_t flags;
	u8 lo = 0xff;
	u8 hi = 0xff;

	local_irq_save(flags);
	cmd_set(CMD_ALE);
	ad_set(addr & 0xff);
	clk_hi();

	cmd_set(CMD_READ);
	ad_free();
	clk_hi();

	cmd_set(CMD_READ);
	if (devsel_get())
		goto out;

	lo = ad_get();
	clk_hi();
	
	cmd_set(CMD_NOP);
	hi = ad_get();
out:
	clk_hi();

	cmd_set(CMD_NOP);
	clk_hi();

	local_irq_restore(flags);

	return (hi << 8) | lo;
}

/**
 * ioread16 - read 16-bit word from I/O memory address @addr
 * @addr: I/O memory address
 */
u16 ioread16_timer(unsigned int addr) __using(TIMER_BANK) __reentrant
{
	cpu_flags_t flags;
	u8 lo = 0xff;
	u8 hi = 0xff;

	local_irq_save(flags);
	cmd_set(CMD_ALE);
	ad_set(addr & 0xff);
	clk_hi();

	cmd_set(CMD_READ);
	ad_free();
	clk_hi();

	cmd_set(CMD_READ);
	if (devsel_get())
		goto out;
	lo = ad_get();
	clk_hi();
	
	cmd_set(CMD_NOP);
	hi = ad_get();
out:
	clk_hi();

	cmd_set(CMD_NOP);
	clk_hi();

	local_irq_restore(flags);

	return (hi << 8) | lo;
}

/**
 * ioread8_copy - read data block from bus
 * @addr: first address
 * @dst: destination address
 * @count: number of bytes to copy
 */
char ioread8_copy(unsigned int addr, u8 __kernel *dst, unsigned char count)
	__reentrant
{
	cpu_flags_t flags;
	u8 c;
	char ret = -EIO;

	/*
	 * This implementation is pipelined and minimal number of bytes is 2.
	 */
	if (count < 2) {
		if (count == 1) {
			/* FIXME: report errors */
			*dst = ioread8(addr);
		}
		return 0;
	}

	local_irq_save(flags);

	cmd_set(CMD_ALE);
	ad_set(addr & 0xff);
	clk_hi();

	/* read first byte */
	cmd_set(CMD_READ);
	ad_free();
	clk_hi();

	/* read next byte */
	cmd_set(CMD_READ);
	clk_hi();

	/* check if any device decoded this address */
	if (!devsel_get()) {
		/* read first byte from device */
		c = ad_get();
		*dst++ = c;
		count -= 2;

		if (count) {
			/* read middle bytes */
			do {
				cmd_set(CMD_READ);
				c = ad_get();
				clk_hi();
				*dst++ = c;
			} while (--count > 0);
		}

		/* read last byte */
		cmd_set(CMD_NOP);
		c = ad_get();
		clk_hi();
		*dst = c;
		ret = 0;
	} else {
		memset_kernel(dst, 0xff, count);
	}

	cmd_set(CMD_NOP);
	clk_hi();

	local_irq_restore(flags);

	return ret;
}

static void beerbus_init(void) __reentrant
{
	devsel_hi();
	ad_free();
	/* setup control signals */
	cmd_set(CMD_NOP);
	clk_hi();

	/* setup bus clock, CPU Clock / 2 using PWM0 */
	/* PWM at pins 2.6 and 2.7 */
	CFG841 &= ~CFG841_PWPO;

	/* PWM freq = XTAL / 1 */
	PWMCON = PWMCON_SNGL | PWMCON_MODE_SINGLE
		| PWMCON_CDIV_1 | PWMCON_CSEL_XTAL;

	/*
	 * 1 / 2 duty, PWM freq / 2
	 *
	 * The output frequency should be 6.67 MHz +/- 50%.
	 * CPU Freq / 2 is a good value, because:
	 * 	- it's maximal possible frequency, for lower
	 * 	CPU frequency is always the best possible
	 * 	- the maximal CPU frequency is 20 MHz, so
	 * 	maximal output frequency will be 10 MHz
	 * 	still in +/- 50%.
	 */
	PWM1L = 1;
	PWM1H = 0;

	PWM0L = 0;
	PWM0H = 0;
//	bus_reset();
}

arch_initcall(beerbus_init)
